"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("./metadata/metadata");
const client_1 = require("./client");
class Redisk {
    constructor(metadata, client) {
        this.metadata = metadata;
        this.client = client;
    }
    static init(options) {
        return new Redisk(new metadata_1.Metadata(), new client_1.RedisClient(options));
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.client.closeConnection();
        });
    }
    getClient() {
        return this.client;
    }
    save(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques, primary, canBeListed, indexes, properties, hasOneRelations } = this.metadata.getEntityMetadataFromInstance(entity);
            const hashKey = name + ':' + entity[primary];
            const persistedEntity = yield this.getOne(entity.constructor, entity[primary]);
            if (persistedEntity !== null) {
                const changedFields = [];
                for (const property of properties) {
                    if (entity[property.name] !== persistedEntity[property.name]) {
                        changedFields.push(property.name);
                        if (entity[property.name] === null) {
                            yield this.client.hdel(hashKey, property.name);
                        }
                        if (hasOneRelations !== undefined && hasOneRelations[property.name] && hasOneRelations[property.name].cascadeUpdate && entity[property.name] !== null) {
                            yield this.save(entity[property.name]);
                        }
                        if (property.searchable) {
                            yield this.client.srem(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + persistedEntity[property.name].toLowerCase());
                        }
                        if (property.sortable) {
                            yield this.client.zrem(this.getSortableKeyName(name, property.name), persistedEntity[property.name]);
                        }
                    }
                }
                if (indexes) {
                    const indexesChanged = changedFields.some(value => indexes.indexOf(value) >= 0);
                    if (indexesChanged) {
                        yield this.dropIndexes(persistedEntity, entity[primary]);
                    }
                }
                if (uniques) {
                    const uniquesChanged = changedFields.some(value => uniques.indexOf(value) >= 0);
                    if (uniquesChanged) {
                        yield this.dropUniqueKeys(persistedEntity);
                    }
                }
            }
            if (uniques) {
                for (const uniqueName of uniques) {
                    const entityWithUnique = yield this.getOne(entity.constructor, entity[uniqueName], uniqueName);
                    if (entityWithUnique !== null && entityWithUnique[primary] !== entity[primary]) {
                        throw new Error(uniqueName + ' is not unique!');
                    }
                    if (entity[uniqueName] !== null) {
                        yield this.client.set(this.getUniqueKeyName(name, uniqueName) + ':' + entity[uniqueName], entity[primary]);
                    }
                }
            }
            const valuesToStore = [];
            for (const property of properties) {
                if (entity[property.name] !== null) {
                    let valueToStore = this.convertPropertyTypeToPrimitive(property, entity[property.name]);
                    if (hasOneRelations !== undefined && hasOneRelations[property.name]) {
                        const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[property.name].entity);
                        valueToStore = entity[property.name][relatedEntity.primary];
                        if (hasOneRelations[property.name].cascadeInsert && persistedEntity === null && entity[property.name] !== null) {
                            yield this.save(entity[property.name]);
                        }
                    }
                    valuesToStore.push(property.name);
                    valuesToStore.push(valueToStore);
                    if (property.sortable === true) {
                        yield this.client.zadd(this.getSortableKeyName(name, property.name), this.convertPropertyTypeToPrimitive(property, entity[property.name]), entity[primary]);
                    }
                    if (property.searchable === true) {
                        yield this.client.sadd(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + entity[property.name].toLowerCase());
                    }
                }
            }
            yield this.client.hmset(hashKey, valuesToStore);
            if (indexes) {
                for (const indexName of indexes) {
                    let value = entity[indexName];
                    if (hasOneRelations !== undefined && hasOneRelations[indexName] && entity[indexName] !== null) {
                        const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[indexName].entity);
                        value = entity[indexName][relatedEntity.primary];
                    }
                    if (value !== null) {
                        yield this.client.sadd(this.getIndexKeyName(name, indexName, value), entity[primary]);
                    }
                }
            }
            if (canBeListed && persistedEntity === null) {
                yield this.client.rpush(this.getListKeyName(name), entity[primary]);
            }
            return null;
        });
    }
    count(entityType) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name } = this.metadata.getEntityMetadataFromType(entityType);
            const keyName = this.getListKeyName(name);
            return yield this.client.llen(keyName);
        });
    }
    list(entityType, limit, offset, orderBy) {
        return __awaiter(this, void 0, void 0, function* () {
            const ids = yield this.listIds(entityType, limit, offset, orderBy);
            const response = [];
            for (const id of ids) {
                response.push(yield this.getOne(entityType, id));
            }
            return response;
        });
    }
    find(entityType, conditions, limit, offset, type = 'AND') {
        return __awaiter(this, void 0, void 0, function* () {
            const ids = yield this.findIds(entityType, conditions, type);
            const response = [];
            if (limit !== undefined || offset !== undefined) {
                if (limit === undefined || offset === undefined) {
                    throw new Error('You must specify limit and offset, not just one arg.');
                }
                for (let index = offset; index < ids.length && index < (limit + offset); index++) {
                    response.push(yield this.getOne(entityType, ids[index]));
                }
            }
            else {
                for (const id of ids) {
                    response.push(yield this.getOne(entityType, id));
                }
            }
            return response;
        });
    }
    search(entityType, condition, limit) {
        return __awaiter(this, void 0, void 0, function* () {
            const ids = yield this.searchIds(entityType, condition, limit);
            const response = [];
            const numberOfResult = (ids.length < limit) ? ids.length : limit;
            for (let index = 0; index < numberOfResult; index++) {
                response.push(yield this.getOne(entityType, ids[index]));
            }
            return response;
        });
    }
    searchIds(entityType, condition, limit) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name } = this.metadata.getEntityMetadataFromType(entityType);
            const key = this.getSearchableKeyName(name, condition.key);
            const value = this.getSearchableValuePrefix('*') + '*' + condition.value.toLowerCase() + '*';
            const response = [];
            let finishedScanning = false;
            let cursor = 0;
            while (!finishedScanning) {
                const scanResponse = (yield this.client.sscan(key, cursor, value));
                cursor = scanResponse.cursor;
                response.push(...scanResponse.data.map((id) => id.match(/.+?(?=\:_id_:)/g)[0]));
                if (cursor === 0 || response.length === limit) {
                    finishedScanning = true;
                }
            }
            return response;
        });
    }
    findIds(entityType, conditions, type = 'AND') {
        return __awaiter(this, void 0, void 0, function* () {
            if (conditions.length === 0) {
                throw new Error('You should at least specify one key to search');
            }
            const { name } = this.metadata.getEntityMetadataFromType(entityType);
            const keyNames = [];
            for (const condition of conditions) {
                keyNames.push(this.getIndexKeyName(name, condition.key, String(condition.value)));
            }
            if (type === 'AND') {
                return yield this.client.sinter(keyNames);
            }
            else {
                return yield this.client.sunion(keyNames);
            }
        });
    }
    listIds(entityType, limit, offset, orderBy) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, canBeListed } = this.metadata.getEntityMetadataFromType(entityType);
            if (!canBeListed) {
                throw new Error(entityType.name + ' can\'t be listed!');
            }
            const keyName = this.getListKeyName(name);
            let start = 0;
            let stop = -1;
            if (offset !== undefined) {
                start = offset;
            }
            if (limit !== undefined) {
                stop = start + limit - 1;
            }
            if (orderBy !== undefined) {
                const sortableKey = this.getSortableKeyName(name, orderBy.field);
                if (orderBy.strategy === 'ASC') {
                    return yield this.client.zrange(sortableKey, start, stop);
                }
                else {
                    return yield this.client.zrevrange(sortableKey, start, stop);
                }
            }
            return yield this.client.lrange(keyName, start, stop);
        });
    }
    delete(entityType, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques, indexes, canBeListed } = this.metadata.getEntityMetadataFromType(entityType);
            const hashKey = name + ':' + id;
            const persistedEntity = yield this.getOne(entityType, id);
            if (uniques) {
                yield this.dropUniqueKeys(persistedEntity);
            }
            if (indexes) {
                yield this.dropIndexes(persistedEntity, id);
            }
            if (canBeListed) {
                yield this.client.lrem(this.getListKeyName(name), 1, id);
            }
            yield this.dropSortables(persistedEntity);
            yield this.dropSearchables(persistedEntity);
            yield this.client.del(hashKey);
        });
    }
    getOne(entityType, value, key) {
        return __awaiter(this, void 0, void 0, function* () {
            const entity = Object.create(entityType.prototype);
            const valueAsString = String(value);
            const { name, uniques, primary, properties, hasOneRelations } = this.metadata.getEntityMetadataFromType(entityType);
            let id;
            if (key !== undefined && key !== primary) {
                let indexKey;
                for (const uniqueName of uniques) {
                    if (uniqueName === key) {
                        indexKey = this.getUniqueKeyName(name, uniqueName);
                    }
                }
                if (indexKey === undefined) {
                    throw new Error(key + ' is not an unique field!');
                }
                id = yield this.client.get(indexKey + ':' + valueAsString);
            }
            else {
                id = valueAsString;
            }
            const hashKey = name + ':' + id;
            const result = yield this.client.hmget(hashKey, properties.map((property) => property.name));
            let index = 0;
            for (const resultKey of result) {
                if (hasOneRelations !== undefined && hasOneRelations[properties[index].name] && resultKey !== null) {
                    entity[properties[index].name] = yield this.getOne(hasOneRelations[properties[index].name].entityType, resultKey);
                }
                else {
                    entity[properties[index].name] = this.convertStringToPropertyType(properties[index], resultKey);
                }
                index++;
            }
            if (entity[primary] === null) {
                return null;
            }
            return entity;
        });
    }
    convertPropertyTypeToPrimitive(property, value) {
        if (property.type === 'Date') {
            return value.valueOf();
        }
        return String(value);
    }
    convertStringToPropertyType(property, value) {
        let convertedValue = value;
        switch (property.type) {
            case 'Boolean':
                convertedValue = value === 'true';
                break;
            case 'Number':
                convertedValue = Number(value);
                break;
            case 'Date':
                convertedValue = new Date(Number(value));
                break;
        }
        return convertedValue;
    }
    dropUniqueKeys(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const uniqueName of uniques) {
                yield this.client.del(this.getUniqueKeyName(name, uniqueName) + ':' + entity[uniqueName]);
            }
        });
    }
    dropIndexes(entity, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, indexes, hasOneRelations } = this.metadata.getEntityMetadataFromInstance(entity);
            if (indexes) {
                for (const indexName of indexes) {
                    let value = entity[indexName];
                    if (hasOneRelations !== undefined && hasOneRelations[indexName]) {
                        const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[indexName].entity);
                        value = entity[indexName][relatedEntity.primary];
                    }
                    yield this.client.srem(this.getIndexKeyName(name, indexName, value), id);
                }
            }
        });
    }
    dropSearchables(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, properties, primary } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const property of properties) {
                if (property.searchable === true) {
                    yield this.client.srem(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + entity[property.name].toLowerCase());
                }
            }
        });
    }
    dropSortables(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, properties, primary } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const property of properties) {
                if (property.sortable === true) {
                    yield this.client.zrem(this.getSortableKeyName(name, property.name), entity[primary]);
                }
            }
        });
    }
    getIndexKeyName(entityName, indexName, indexValue) {
        return entityName + ':index:' + indexName + ':' + indexValue;
    }
    getListKeyName(entityName) {
        return entityName + ':list';
    }
    getUniqueKeyName(entityName, uniqueName) {
        return entityName + ':unique:' + uniqueName;
    }
    getSortableKeyName(entityName, fieldName) {
        return entityName + ':sort:' + fieldName;
    }
    getSearchableKeyName(entityName, fieldName) {
        return entityName + ':search:' + fieldName;
    }
    getSearchableValuePrefix(id) {
        return id + ':_id_:';
    }
}
exports.Redisk = Redisk;
